# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from collections import OrderedDict

from lewis.core.processor import CanProcess
from lewis.core.statemachine import StateMachine
from lewis.devices import StateMachineDevice

from . import states
from .bearings import MagneticBearings


class SimulatedBearings(CanProcess, MagneticBearings):
    def __init__(self) -> None:
        super(SimulatedBearings, self).__init__()

        self._csm = StateMachine(
            {
                "initial": "resting",
                "transitions": {
                    ("resting", "levitating"): lambda: self._levitate,
                    ("levitating", "levitated"): self.levitationComplete,
                    ("levitated", "delevitating"): lambda: not self._levitate,
                    ("delevitating", "resting"): self.delevitationComplete,
                },
            }
        )

        self._levitate = False

    def engage(self) -> None:
        self.levitate()

    def disengage(self) -> None:
        self.delevitate()

    def levitate(self) -> None:
        self._levitate = True

    def delevitate(self) -> None:
        self._levitate = False

    def levitationComplete(self) -> bool:
        return True

    def delevitationComplete(self) -> bool:
        return True

    def doProcess(self, dt) -> None:
        self._csm.process(dt)

    @property
    def ready(self):
        return self._csm.state == "levitated" and self._levitate

    @property
    def idle(self):
        return self._csm.state == "resting" and not self._levitate


class SimulatedChopper(StateMachineDevice):
    _bearings = None

    def _initialize_data(self) -> None:
        self.speed = 0.0
        self.target_speed = 0.0

        self.phase = 0.0
        self.target_phase = 0.0

        self.parking_position = 0.0
        self.target_parking_position = 0.0
        self.auto_park = False

        self._park_commanded = False
        self._stop_commanded = False
        self._start_commanded = False
        self._idle_commanded = False
        self._phase_commanded = False
        self._shutdown_commanded = False
        self._initialized = False

        if self._bearings is None:
            self._bearings = SimulatedBearings()

    def _get_state_handlers(self):
        return {
            "init": states.DefaultInitState(),
            "bearings": {"in_state": self._bearings},
            "stopped": states.DefaultStoppedState(),
            "stopping": states.DefaultStoppingState(),
            "accelerating": states.DefaultAcceleratingState(),
            "phase_locking": states.DefaultPhaseLockingState(),
            "phase_locked": states.DefaultPhaseLockedState(),
            "idle": states.DefaultIdleState(),
            "parking": states.DefaultParkingState(),
            "parked": states.DefaultParkedState(),
        }

    def _get_initial_state(self) -> str:
        return "init"

    def _get_transition_handlers(self):
        return OrderedDict(
            [
                (("init", "bearings"), lambda: self.initialized),
                (("bearings", "stopped"), lambda: self._bearings.ready),
                (("bearings", "init"), lambda: self._bearings.idle),
                (
                    ("parking", "parked"),
                    lambda: self.parking_position == self.target_parking_position,
                ),
                (("parking", "stopping"), lambda: self._stop_commanded),
                (("parked", "stopping"), lambda: self._stop_commanded),
                (("parked", "accelerating"), lambda: self._start_commanded),
                (("stopped", "accelerating"), lambda: self._start_commanded),
                (("stopped", "parking"), lambda: self._park_commanded),
                (("stopped", "bearings"), lambda: self._shutdown_commanded),
                (("accelerating", "stopping"), lambda: self._stop_commanded),
                (("accelerating", "idle"), lambda: self._idle_commanded),
                (
                    ("accelerating", "phase_locking"),
                    lambda: self.speed == self.target_speed,
                ),
                (("idle", "accelerating"), lambda: self._start_commanded),
                (("idle", "stopping"), lambda: self._stop_commanded),
                (("phase_locking", "stopping"), lambda: self._stop_commanded),
                (
                    ("phase_locking", "phase_locked"),
                    lambda: self.phase == self.target_phase,
                ),
                (("phase_locking", "idle"), lambda: self._idle_commanded),
                (("phase_locked", "accelerating"), lambda: self._start_commanded),
                (("phase_locked", "phase_locking"), lambda: self._phase_commanded),
                (("phase_locked", "stopping"), lambda: self._stop_commanded),
                (("phase_locked", "idle"), lambda: self._idle_commanded),
                (("stopping", "accelerating"), lambda: self._start_commanded),
                (("stopping", "stopped"), lambda: self.speed == 0.0),
                (("stopping", "idle"), lambda: self._idle_commanded),
            ]
        )

    @property
    def state(self):
        """
        The current state of the chopper. This parameter is read-only, it is
        determined by the internal state machine of the device.
        """
        return self._csm.state

    @property
    def initialized(self):
        return self._initialized

    def initialize(self) -> None:
        if self._csm.can("bearings") and not self.initialized:
            self._initialized = True
            self._bearings.engage()

    def deinitialize(self) -> None:
        if self._csm.can("bearings") and self.initialized:
            self._shutdown_commanded = True
            self._bearings.disengage()

    def park(self) -> None:
        if self._csm.can("parking"):
            self._park_commanded = True

    @property
    def parked(self):
        return self._csm.state == "parked"

    def stop(self) -> None:
        if self._csm.can("stopping"):
            self._stop_commanded = True

    @property
    def stopped(self):
        return self._csm.state == "stopped"

    def start(self) -> None:
        if self._csm.can("accelerating") and self.target_speed > 0.0:
            self._start_commanded = True
        else:
            self.stop()

    @property
    def started(self):
        return self._csm.state == "accelerating"

    def unlock(self) -> None:
        if self._csm.can("idle"):
            self._idle_commanded = True

    @property
    def idle(self):
        return self._csm.state == "idle"

    def lock_phase(self) -> None:
        if self._csm.can("phase_locking"):
            self._phase_commanded = True

    @property
    def phase_locked(self):
        return self._csm.state == "phase_locked"



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains :class:`DeviceBase` as a base class for other device classes and
infrastructure that can import devices from a module (:class:`DeviceRegistry`). The latter also
produces factory-like objects that create device instances and interfaces based on setups
(:class:`DeviceBuilder`).
"""

import importlib
from typing import NoReturn

from lewis.core.exceptions import LewisException
from lewis.core.logging import has_log
from lewis.core.utils import get_members, get_submodules


@has_log
class DeviceBase:
    """
    This class is a common base for :class:`~lewis.devices.Device` and
    :class:`~lewis.devices.StateMachineDevice`. It is mainly used in the device
    discovery process.
    """


@has_log
class InterfaceBase:
    """
    This class is a common base for protocol specific interfaces that are exposed by a subclass of
    :class:`~lewis.core.adapters.Adapter`. This base class is not meant to be used directly in
    a device package - this is what the interfaces in :mod:`lewis.adapters` are for.

    There is a 1:1 correspondence between device and interface, where the interface holds a
    reference to the device. It can be changed through the ``device``-property.
    """

    protocol = None

    def __init__(self) -> None:
        super(InterfaceBase, self).__init__()
        self._device = None

    @property
    def adapter(self) -> NoReturn:
        """
        Adapter type that is required to process and expose interfaces of this type. Must be
        implemented in subclasses.
        """
        raise NotImplementedError(
            "An interface type must specify which adapter it is compatible "
            "with. Please implement the adapter-property."
        )

    @property
    def device(self):
        """
        The device this interface is bound to. When a new device is set, :meth:`_bind_device` is
        called, where the interface can react to the device change if necessary.
        """
        return self._device

    @device.setter
    def device(self, new_device) -> None:
        self._device = new_device
        self._bind_device()

    def _bind_device(self) -> None:
        """
        This method should perform any binding steps between device and interface. The result
        of this binding step is generally used by the adapter to process network traffic.

        The default implementation does nothing.
        """
        pass


def is_device(obj):
    """
    Returns True if obj is a device type (derived from DeviceBase), but not defined in
    :mod:`lewis.core.devices` or :mod:`lewis.devices`.

    :param obj: Object to test.
    :return: True if obj is a device type.
    """
    return (
        isinstance(obj, type)
        and issubclass(obj, DeviceBase)
        and obj.__module__ not in ("lewis.devices", "lewis.core.devices")
    )


def is_interface(obj):
    """
    Returns True if obj is an interface (derived from :class:`InterfaceBase`), but not defined in
    :mod:`lewis.adapters`, where concrete interfaces for protocols are defined.

    :param obj: Object to test.
    :return: True if obj is an interface type.
    """
    return (
        isinstance(obj, type)
        and issubclass(obj, InterfaceBase)
        and not (
            obj.__module__.startswith("lewis.core.devices")
            or obj.__module__.startswith("lewis.adapters")
        )
    )


@has_log
class DeviceBuilder:
    """
    This class takes a module object (for example imported via importlib.import_module or via the
    :class:`DeviceRegistry`) and inspects it so that it's possible to construct devices and
    interfaces.

    In order for the class to work properly, the device module has to adhere to a few rules.
    Device types, which means classes inheriting from :class:`DeviceBase`, are imported directly
    from the device module, equivalent to the following:

    .. sourcecode :: Python

        from device_name import SimulatedDeviceType

    If ``SimulatedDeviceType`` is defined in the ``__init__.py``, there's nothing else to do. If
    the device class is defined elsewhere, it must be imported in the ``__init__.py`` file as
    written above. If there is only one device type (which is probably the most common case), it is
    assumed to be default device type.

    Setups are discovered in two locations, the first one is a dict called ``setups`` in the device
    module, which must contain setup names as keys and as values again a dict. This inner dict has
    one mandatory key called ``device_type`` and one optional key ``parameters`` containing the
    constructor arguments for the specified device type:

    .. sourcecode:: Python

        setups = dict(
            broken=dict(
                device_type=SimulatedDeviceType,
                parameters=dict(
                    override_initial_state="error",
                    override_initial_data=dict(target=-10, position=-20.0),
                ),
            )
        )

    The other location is a sub-package called `setups`, which should in turn contain modules. Each
    module must contain a variable ``device_type`` and a variable ``parameters`` which are
    analogous to the keys in the dict described above. This allows for more complex setups which
    define additional classes and so on.

    The ``default`` setup is special, it is used when no setup is supplied to
    :meth:`create_device`. If the setup ``default`` is not defined, one is created with the default
    device type. This has two consequences, no setups need to be defined for very simple devices,
    but if multiple device types are defined, a ``default`` setup must be defined.

    A setup can be supplied to the :meth:`create_device`.

    Lastly, the builder tries to discover device interfaces, which are currently classes based on
    :class:`lewis.adapters.InterfaceBase`. These are looked for in the module and in a sub-package
    called ``interfaces`` (which should contain modules with adapters like the ``setups`` package).

    Each interface has a protocol, if a protocol occurs more than once in a device module,
    a RuntimeError is raised.
    """

    def __init__(self, module) -> None:
        self._module = module

        submodules = get_submodules(self._module)

        self._device_types = self._discover_devices(submodules.get("devices"))
        self._setups = self._discover_setups(submodules.get("setups"))
        self._interfaces = self._discover_interfaces(submodules.get("interfaces"))

        self.log.debug(
            "Discovered the following items in '%s': Devices: %s; Setups: %s; Interfaces: %s",
            self._module.__name__,
            ", ".join(device_t.__name__ for device_t in self._device_types),
            ", ".join(self._setups.keys()),
            ", ".join("(%s: %s)" % (k, v.__name__) for k, v in self._interfaces.items()),
        )

    def _discover_devices(self, devices_package):
        devices = list(get_members(self._module, is_device).values())

        if devices_package is None:
            return devices

        for module in get_submodules(devices_package).values():
            devices += list(get_members(module, is_device).values())

        return devices

    def _discover_setups(self, setups_package):
        setups = getattr(self._module, "setups", {})

        all_setups = setups if isinstance(setups, dict) else {}

        if setups_package is not None:
            for name, setup_module in get_submodules(setups_package).items():
                existing_setup = all_setups.get(name)

                if existing_setup is not None:
                    raise RuntimeError(
                        "The setup '{}' is defined twice in device '{}'.".format(
                            existing_setup, self.name
                        )
                    )

                all_setups[name] = {
                    "device_type": getattr(setup_module, "device_type", self.default_device_type),
                    "parameters": getattr(setup_module, "parameters", {}),
                }

        if "default" not in all_setups:
            all_setups["default"] = {"device_type": self.default_device_type}

        return all_setups

    def _discover_interfaces(self, interface_package):
        all_interfaces = []

        if interface_package is not None:
            for interface_module in get_submodules(interface_package).values():
                all_interfaces += list(get_members(interface_module, is_interface).values())

        all_interfaces += list(get_members(self._module, is_interface).values())

        interfaces = {}
        for interface in all_interfaces:
            existing_interface = interfaces.get(interface.protocol)

            if existing_interface is not None:
                raise RuntimeError(
                    "The protocol '{}' is defined in two interfaces for device '{}':\n"
                    "    {} (in {})\n"
                    "    {} (in {})\n"
                    "One of the protocol names needs to be changed.".format(
                        interface.protocol,
                        self.name,
                        existing_interface.__name__,
                        existing_interface.__module__,
                        interface.__name__,
                        interface.__module__,
                    )
                )

            interfaces[interface.protocol] = interface

        return interfaces

    @property
    def framework_version(self):
        return getattr(self._module, "framework_version", None)

    @property
    def name(self):
        """
        The name of the device, which is also the name of the device module.
        """
        return self._module.__name__.split(".")[-1]

    @property
    def device_types(self):
        """
        This property contains a dict of all device types in the device module. The keys are
        type names, the values are the types themselves.
        """
        return self._device_types

    @property
    def default_device_type(self):
        """
        If the module only defines one device type, it is the default device type. It is used
        whenever a setup does not provide a ``device_type``.
        """
        if len(self.device_types) == 1:
            return self.device_types[0]

        return None

    @property
    def interfaces(self):
        """
        This property contains a map with protocols as keys and interface types as values.
        The types are imported from the ``interfaces`` sub-module and from the device module
        itself. If two interfaces with the same protocol are discovered, a RuntimeError is raiesed.
        """
        return self._interfaces

    @property
    def protocols(self):
        """All available protocols for this device."""
        return list(self.interfaces.keys())

    @property
    def default_protocol(self):
        """In case only one protocol exists for the device, this is the default protocol."""
        if len(self.protocols) == 1:
            return self.protocols[0]

        return None

    @property
    def setups(self):
        """
        A map with all available setups. Setups are imported from the ``setups`` dictionary
        in a device module and from the ``setups`` sub-module. If no ``default``-setup exists,
        one is created using the default_device_type. If there are several device types in
        the module, the default setup must be provided explicitly.
        """
        return self._setups

    def _create_device_instance(self, device_type, **kwargs):
        if device_type not in self.device_types:
            raise RuntimeError("Can not create instance of non-device type.")

        return device_type(**kwargs)

    def create_device(self, setup=None):
        """
        Creates a device object according to the provided setup. If no setup is provided,
        the default setup is used. If the setup can't be found, a LewisException is raised.
        This can also happen if the device type specified in the setup is invalid.

        :param setup: Name of the setup from which to create device.
        :return: Device object initialized according to the provided setup.
        """
        setup_name = setup if setup is not None else "default"

        if setup_name not in self.setups:
            raise LewisException(
                "Failed to find setup '{}' for device '{}'. "
                "Available setups are:\n    {}".format(
                    setup, self.name, "\n    ".join(self.setups.keys())
                )
            )

        setup_data = self.setups[setup_name]
        device_type = setup_data.get("device_type") or self.default_device_type

        self.log.debug(
            "Trying to create device '%s' (setup: %s, device type: %s)",
            self.name,
            setup_name,
            device_type.__name__ if device_type else "",
        )

        try:
            return self._create_device_instance(device_type, **setup_data.get("parameters", {}))
        except RuntimeError:
            raise LewisException(
                "The setup '{}' you tried to load does not specify a valid device type, but the "
                "device module '{}' provides multiple device types so that no meaningful "
                "default can be deduced.".format(setup_name, self.name)
            )

    def get_interface_type(self, protocol=None):
        return self.interfaces[protocol]

    def create_interface(self, protocol=None, *args, **kwargs):
        """
        Returns an interface that implements the provided protocol. If the protocol is not
        known, a LewisException is raised. All additional arguments are forwarded
        to the interface constructor (see :class:`~lewis.adapters.Adapter` for details).

        :param protocol: Protocol which the interface must implement.
        :param args: Positional arguments that are passed on to the interface.
        :param kwargs: Keyword arguments that are passed on to the interface.
        :return: Instance of the interface type.
        """
        protocol = protocol if protocol is not None else self.default_protocol

        self.log.debug("Trying to create interface for protocol '%s'", protocol)

        try:
            return self.interfaces[protocol](*args, **kwargs)
        except KeyError:
            raise LewisException(
                "'{}' is not a valid protocol for device '{}', select one via the -p option.\n"
                "Available protocols are: \n    {}".format(
                    protocol, self.name, "\n    ".join(self.interfaces.keys())
                )
            )


@has_log
class DeviceRegistry:
    """
    This class takes the name of a module and constructs a :class:`DeviceBuilder` from
    each sub-module. The available devices can be queried and a DeviceBuilder can be
    obtained for each device:

    .. sourcecode:: Python

        from lewis.core.devices import DeviceRegistry

        registry = DeviceRegistry("lewis.devices")
        chopper_builder = registry.device_builder("chopper")

        # construct device, interface, ...

    If the module can not be imported, a LewisException is raised.

    :param device_module: Name of device module from which devices are loaded.
    """

    def __init__(self, device_module) -> None:
        try:
            self._device_module = importlib.import_module(device_module)
        except ImportError:
            raise LewisException(
                "Failed to import module '{}' for device discovery. "
                "Make sure that it is in the PYTHONPATH.\n"
                "See also the -a option of lewis.".format(device_module)
            )

        self._devices = {
            name: DeviceBuilder(module)
            for name, module in get_submodules(self._device_module).items()
        }

        self.log.debug(
            "Devices loaded from '%s': %s",
            device_module,
            ", ".join(self._devices.keys()),
        )

    @property
    def devices(self):
        """All available device names."""
        return self._devices.keys()

    def device_builder(self, name):
        """
        Returns a :class:`DeviceBuilder` instance that can be used to create device objects
        based on setups, as well as device interfaces. If the device name is not stored
        in the internal map, a LewisException is raised.

        Each DeviceBuilder has a ``framework_version``-member, which specifies the version
        of Lewis the device has been written for. If the version does not match the current
        framework version, it is only possible to obtain those device builders calling the
        method with ``strict_versions`` set to ``False``, otherwise a
        :class:`~lewis.core.exceptions.LewisException` is raised. A warning message is logged
        in all cases. If ``framework_version`` is ``None`` (e.g. not specified at all), it
        is accepted unless ``strict_versions`` is set to ``True``.

        :param name: Name of the device.
        :return: :class:`DeviceBuilder`-object for requested device.
        """
        try:
            return self._devices[name]
        except KeyError:
            raise LewisException(
                "No device with the name '{}' could be found. "
                "Possible names are:\n    {}\n"
                "See also the -k option to add inspect a different module.".format(
                    name, "\n    ".join(self.devices)
                )
            )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains classes to expose objects via a JSON-RPC over ZMQ server. Lewis uses
this infrastructure in :class:`~lewis.core.simulation.Simulation`.

.. seealso::

    Client classes for the service defined in this module can be found in
    :mod:`~lewis.core.control_client`.

"""

import inspect
import json
import socket

import zmq
from jsonrpc import JSONRPCResponseManager

from lewis.core.exceptions import LewisException
from lewis.core.logging import has_log


class ExposedObject:
    """
    ExposedObject is a class that makes it easy to expose an object via the
    JSONRPCResponseManager from the json-rpc package, where it can serve as a dispatcher.
    For this purpose it exposes a read-only dict-like interface.

    The basic problem solved by this wrapper is that plain data members of an object are not
    really captured well by the RPC-approach, where a client performs function calls on a
    remote machine and gets the result back.

    The supplied object is inspected using dir(object) and all entries that do not start
    with a _ are exposed in a way that depends on whether the corresponding member
    is a method or a property (either in the Python-sense or the general OO-sense). Methods
    are stored directly, and stored in an internal dict where the method name is the key and
    the callable method object is the value. For properties, a getter- and a setter function
    are generated, which are then stored in the same dict. The names of these methods for
    a property called ``a`` are ``a:get`` and ``a:set``. The separator has been chosen to be
    colon because it can't be part of a valid Python identifier.

    If the second argument is not empty, it is interpreted to be the list of members
    to expose and only those are actually exposed. This can be used to explicitly expose
    members of an object that start with an underscore. If all but one or two members
    should be exposed, it's also possible to use the exclude-argument to explicitly
    exclude a few members. Both parameters can be used in combination, the exclude-list
    takes precedence.

    In certain situations it is desirable to acquire a lock before accessing the exposed object,
    for example when multiple threads are accessing it on the server side. For this purpose,
    the ``lock``-parameter can be used. If it is not ``None``, the exposed methods are wrapped
    in a function that acquires the lock before accessing ``obj``, and releases it afterwards.

    :param obj: The object to expose.
    :param members: This list of methods will be exposed. (defaults to all public members)
    :param exclude: Members in this list will not be exposed.
    :param exclude_inherited: Should inherited members be excluded? (defaults to False)
    :param lock: ``threading.Lock`` that is used when accessing ``obj``.
    """

    def __init__(self, obj, members=None, exclude=None, exclude_inherited=False, lock=None) -> None:
        super(ExposedObject, self).__init__()

        self._object = obj
        self._function_map = {}
        self._lock = lock

        self._add_function(":api", self.get_api)

        exposed_members = members if members else self._public_members()
        exclude = list(exclude or [])
        if exclude_inherited:
            for base in inspect.getmro(type(obj))[1:]:
                exclude += dir(base)

        for method in exposed_members:
            if method not in exclude:
                self._add_member_wrappers(method)

    def _public_members(self):
        """
        Returns a list of members that do not start with an underscore.
        """
        return [prop for prop in dir(self._object) if not prop.startswith("_")]

    def _add_member_wrappers(self, member) -> None:
        """
        This method probes the supplied member of the wrapped object and inserts an appropriate
        entry into the internal method map. Getters and setters for properties get a suffix
        ':get' and ':set' respectively.

        :param member: The member of the wrapped object to expose
        """
        method_object = getattr(type(self._object), member, None) or getattr(self._object, member)

        if callable(method_object):
            self._add_function(member, getattr(self._object, member))
        else:
            self._add_property(member)

    def get_api(self):
        """
        This method returns the class name and a list of exposed methods.
        It is exposed to RPC-clients by an instance of ExposedObjectCollection.

        :return: A dictionary describing the exposed API (consisting of a class name and methods).
        """
        return {
            "class": type(self._object).__name__,
            "methods": list(self._function_map.keys()),
        }

    def __getitem__(self, item):
        return self._function_map[item]

    def __len__(self) -> int:
        return len(self._function_map)

    def __iter__(self):
        return iter(self._function_map)

    def __contains__(self, item) -> bool:
        return item in self._function_map

    def _add_property(self, name) -> None:
        self._add_function("{}:get".format(name), lambda: getattr(self._object, name))
        self._add_function("{}:set".format(name), lambda value: setattr(self._object, name, value))

    def _add_function(self, name, function) -> None:
        if not callable(function):
            raise TypeError("Only callable objects can be exposed.")

        if self._lock is not None:

            def create_locking_wrapper(f):
                def locking_wrapper_function(*args, **kwargs):
                    with self._lock:
                        return f(*args, **kwargs)

                return locking_wrapper_function

            function = create_locking_wrapper(function)

        self._function_map[name] = function

    def _remove_function(self, name) -> None:
        del self._function_map[name]


class ExposedObjectCollection(ExposedObject):
    """
    This class helps expose a number of objects (plain or RPCObject) by
    exposing the methods of each object as

    .. sourcecode:: Python

        name.method

    Furthermore it exposes each object's API as a method with the following name:

    .. sourcecode:: Python

        name: api

    A list of exposed objects can be obtained by calling the following method from the client:

    .. sourcecode:: Python

        :objects

    :param named_objects: Dictionary of of name: object pairs.
    """

    def __init__(self, named_objects) -> None:
        super(ExposedObjectCollection, self).__init__(self, ("get_objects",))
        self._object_map = {}

        if named_objects:
            for name, obj in named_objects.items():
                self.add_object(obj, name)

        self._add_function("get_objects", self.get_objects)

    def add_object(self, obj, name) -> None:
        """
        Adds the supplied object to the collection under the supplied name. If the name is already
        in use, a RuntimeError is raised. If the object is not an instance of
        :class:`ExposedObject`, the method automatically constructs one.

        :param obj: Object to add to the collection.
        :param name: Name of the exposed object.
        """
        if name in self._object_map:
            raise RuntimeError("An object is already registered under that name.")

        exposed_object = self._object_map[name] = (
            obj if isinstance(obj, ExposedObject) else ExposedObject(obj)
        )

        for method_name in exposed_object:
            glue = "." if not method_name.startswith(":") else ""
            self._add_function(name + glue + method_name, exposed_object[method_name])

    def remove_object(self, name) -> None:
        """
        Remove the object exposed under that name. If no object is registered under the supplied
        name, a RuntimeError is raised.

        :param name: Name of object to be removed.
        """
        if name not in self._object_map:
            raise RuntimeError("No object with name {} is registered.".format(name))

        for fn_name in list(self._function_map.keys()):
            if fn_name.startswith(name + ".") or fn_name.startswith(name + ":"):
                self._remove_function(fn_name)

        del self._object_map[name]

    def get_objects(self):
        """Returns the names of the exposed objects."""
        return list(self._object_map.keys())


@has_log
class ControlServer:
    """
    This server opens a ZMQ REP-socket at the given host and port when start_server
    is called.

    The server constructs an :class:`ExposedObjectCollection` from the supplied
    name: object-dictionary and uses that as a handler for JSON-RPC requests. If it is an
    instance of :class:`ExposedObject`, that is used directly.

    Each time process is called, the server tries to get request data and responds to that.
    If there is no data, the method does nothing.

    Please note that this RPC-service comes without any security, authentication, etc.
    Only use it to expose objects on a trusted network and be aware that anyone on that
    network can access the exposed objects without any restrictions.

    :param object_map: Dictionary with name: object-pairs to construct an
                       ExposedObjectCollection or ExposedObject
    :param connection_string: String with host:port pair for binding control server.
    """

    def __init__(self, object_map, connection_string) -> None:
        super(ControlServer, self).__init__()

        try:
            host, port = connection_string.split(":")
        except ValueError:
            raise LewisException(
                "'{}' is not a valid control server initialization string. "
                'A string of the form "host:port" is expected.'.format(connection_string)
            )

        try:
            self.host = socket.gethostbyname(host)
        except socket.gaierror:
            raise LewisException("Could not resolve control server host: {}".format(host))

        self.port = port

        if isinstance(object_map, ExposedObject):
            self._exposed_object = object_map
        else:
            self._exposed_object = ExposedObjectCollection(object_map)

        self._socket = None

    @property
    def is_running(self):
        """
        This property is ``True`` if the server is running.
        """
        return self._socket is not None

    @property
    def exposed_object(self):
        """
        The exposed object. This is a read only property.
        """
        return self._exposed_object

    def start_server(self) -> None:
        """
        Binds the server to the configured host and port and starts listening.
        """
        if self._socket is None:
            context = zmq.Context()
            self._socket = context.socket(zmq.REP)
            self._socket.setsockopt(zmq.RCVTIMEO, 100)
            self._socket.bind("tcp://{0}:{1}".format(self.host, self.port))

            self.log.info("Listening on %s:%s", self.host, self.port)

    def _unhandled_exception_response(self, request_id, exception):
        return {
            "jsonrpc": "2.0",
            "id": request_id,
            "error": {
                "message": "Server error",
                "code": -32000,
                "data": {
                    "message": exception.args,
                    "args": [exception.args],
                    "type": type(exception).__name__,
                },
            },
        }

    def process(self, blocking=False) -> None:
        """
        Each time this method is called, the socket tries to retrieve data and passes
        it to the JSONRPCResponseManager, which in turn passes the RPC to the
        ExposedObjectCollection.

        In case no data are available, the method does nothing. This behavior is required for
        Lewis where everything is running in one thread. The central loop can call process
        at some point to process remote calls, so the RPC-server does not introduce its own
        infinite processing loop.

        If the server has not been started yet (via :meth:`start_server`), a RuntimeError
        is raised.

        :param blocking: If True, this function will block until it has received data or a timeout
                         is triggered. Default is False to preserve behavior of prior versions.
        """
        if self._socket is None:
            raise RuntimeError("The server has not been started yet, use start_server to do so.")

        try:
            request = self._socket.recv_unicode(flags=zmq.NOBLOCK if not blocking else 0)

            self.log.debug("Got request %s", request)

            try:
                response = JSONRPCResponseManager.handle(request, self._exposed_object)
                self._socket.send_unicode(response.json)

                self.log.debug("Sent response %s", response.json)
            except TypeError as e:
                self._socket.send_json(
                    self._unhandled_exception_response(json.loads(request)["id"], e)
                )
        except zmq.Again:
            pass



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.modbus import ModbusBasicDataBank, ModbusInterface
from lewis.devices import Device


class ModbusDevice(Device):
    pass


class ExampleModbusInterface(ModbusInterface):
    """
    The class attributes di, co, ir and hr represent Discrete Inputs, Coils, Input Registers and
    Holding Registers, respectively. Each attribute should be assigned a ModbusDataBank instance
    by the Interface implementation.

    Here, two basic ModbusDataBanks are created and initialized to a default value across the full
    range of valid addresses. One DataBank is shared by di and co, and the other by ir and hr to
    demonstrate overlaid memory segments. If you want each segment to have its own memory, just
    create separate instances for all four.
    """

    di = ModbusBasicDataBank(False)
    co = di
    ir = ModbusBasicDataBank(0)
    hr = ir



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module defines two classes related to one of lewis' essential concepts, namely
the cycle-based approach. :class:`CanProcess` and :class:`CanProcessComposite` implement the
composite design pattern so that it's possible to form a tree of objects which can perform
calculations based on an elapsed time Δt.
"""


class CanProcess:
    """
    The CanProcess class is meant as a base for all things that
    are able to process on the basis of a time delta (dt).

    The base implementation does nothing.

    There are three methods that can be implemented by sub-classes and are called in the
    process-method in this order:

        1. doBeforeProcess
        2. doProcess
        3. doAfterProcess

    The doBefore- and doAfterProcess methods are only called if a doProcess-method exists.
    """

    def __init__(self) -> None:
        super(CanProcess, self).__init__()

    def __call__(self, dt=0):
        self.process(dt)

    def process(self, dt=0) -> None:
        if hasattr(self, "doProcess"):
            if hasattr(self, "doBeforeProcess"):
                self.doBeforeProcess(dt)

            self.doProcess(dt)

            if hasattr(self, "doAfterProcess"):
                self.doAfterProcess(dt)


class CanProcessComposite(CanProcess):
    """
    This subclass of CanProcess is a convenient way of collecting
    multiple items that implement the CanProcess interface.

    Items can be added to the composite like this:

    .. sourcecode:: Python

        composite = CanProcessComposite()
        composite.add_processor(item_that_implements_CanProcess)

    The process-method calls the process-method of each contained
    item. Specific things that have to be done before or after the
    containing items are processed can be implemented in the doBefore-
    and doAfterProcess methods.
    """

    def __init__(self, iterable=()) -> None:
        super(CanProcessComposite, self).__init__()

        self._processors = []

        for item in iterable:
            self.add_processor(item)

    def add_processor(self, other) -> None:
        if isinstance(other, CanProcess):
            self._append_processor(other)

    def _append_processor(self, processor) -> None:
        self._processors.append(processor)

    def doProcess(self, dt) -> None:
        for processor in self._processors:
            processor.process(dt)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from lewis.devices.linkam_t95.devices.device import SimulatedLinkamT95
from lewis.devices.linkam_t95.devices.states import DefaultStartedState
from lewis.devices.linkam_t95.interfaces.stream_interface import (
    LinkamT95StreamInterface,
)

from .utils import assertRaisesNothing


class TestSimulatedLinkamT95(unittest.TestCase):
    def test_default_construction(self):
        assertRaisesNothing(self, SimulatedLinkamT95)

    def test_state_override_construction(self):
        assertRaisesNothing(
            self, SimulatedLinkamT95, override_states={"started": DefaultStartedState()}
        )

    def test_transition_override_construction(self):
        assertRaisesNothing(
            self,
            SimulatedLinkamT95,
            override_transitions={("init", "stopped"): lambda: True},
        )

    def test_default_status(self):
        linkam = LinkamT95StreamInterface()
        linkam.device = SimulatedLinkamT95()

        status_bytes = linkam.get_status()

        self.assertEqual(len(status_bytes), 10, "Byte array should always be 10 bytes")
        self.assertFalse(b"\x00" in status_bytes, "Byte array may not contain zeroes")
        self.assertEqual(0x01, status_bytes[0], "Status byte should be 1 on startup")
        self.assertEqual(0x80, status_bytes[1], "No error flags should be set on startup")
        self.assertEqual(0x80, status_bytes[2], "The pump should not be active on startup")
        self.assertEqual(b"00f0", status_bytes[6:10], "Starting temperature 24C")

    def test_simple_heat(self):
        linkam = LinkamT95StreamInterface()

        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Set up to heat from 24.0 C to 44.0 C at 20.00 C/min
        linkam.set_rate("2000")
        linkam.set_limit("440")
        linkam.start()
        linkam_device.process()

        # Heat for almost a minute but not quite
        linkam_device.process(59.5)
        status_bytes = linkam.get_status()
        self.assertEqual(0x10, status_bytes[0], "Heating status set")
        self.assertNotEqual(status_bytes[6:10], b"01b8")  # Temp != 44.0 C

        # Finish off heating (and overshoot a bit)
        linkam_device.process(5)
        status_bytes = linkam.get_status()
        self.assertEqual(status_bytes[6:10], b"01b8")  # Temp == 44.0 C

        # Should hold now, so temperature should not change
        linkam_device.process(10)
        status_bytes = linkam.get_status()
        self.assertEqual(0x30, status_bytes[0], "Auto-holding at limit")
        self.assertEqual(status_bytes[6:10], b"01b8")  # Temp == 44.0 C

    def test_simple_cool(self):
        linkam = LinkamT95StreamInterface()
        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Set up to cool from 24.0 C to 4.0 C at 20.00 C/min
        linkam.set_rate("2000")
        linkam.set_limit("40")
        linkam.start()
        linkam_device.process()

        # Cool for almost a minute but not quite
        linkam_device.process(59.5)
        status_bytes = linkam.get_status()
        self.assertEqual(0x20, status_bytes[0], "Cooling status set")
        self.assertNotEqual(status_bytes[6:10], b"0028")  # Temp != 4.0 C

        # Finish off cooling (and overshoot a bit)
        linkam_device.process(5)
        status_bytes = linkam.get_status()
        self.assertEqual(status_bytes[6:10], b"0028")  # Temp == 4.0 C

        # Should hold now, so temperature should not change
        linkam_device.process(10)
        status_bytes = linkam.get_status()
        self.assertEqual(0x30, status_bytes[0], "Auto-holding at limit")
        self.assertEqual(status_bytes[6:10], b"0028")  # Temp == 4.0 C

    def test_error_flag_overcool(self):
        linkam = LinkamT95StreamInterface()
        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Ensure flag is not set
        status_bytes = linkam.get_status()
        self.assertFalse(status_bytes[1] & 0x01)

        # Set up to cool from 24.0 C to 4.0 C at 51.00 C/min
        linkam.set_rate("5100")
        linkam.set_limit("40")
        linkam.start()
        linkam_device.process()

        # Ensure flag is set after running a bit
        linkam_device.process(0.1)
        status_bytes = linkam.get_status()
        self.assertTrue(status_bytes[1] & 0x01)

    def test_stop_command(self):
        linkam = LinkamT95StreamInterface()
        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Set up to heat from 24.0 C to 44.0 C at 20.00 C/min
        linkam.set_rate("2000")
        linkam.set_limit("440")
        linkam.start()
        linkam_device.process()

        # Process for some time and then stop
        linkam_device.process(10)
        linkam.stop()
        linkam_device.process()

        # Ensure status byte reports stopped
        status_bytes = linkam.get_status()
        self.assertEqual(status_bytes[0], 0x01)

    def test_hold_and_resume(self):
        linkam = LinkamT95StreamInterface()
        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Set up to cool from 24.0 C to 4.0 C at 20.00 C/min
        linkam.set_rate("2000")
        linkam.set_limit("40")
        linkam.start()
        linkam_device.process()

        # Cool for a while
        linkam_device.process(30)
        status_bytes = linkam.get_status()
        self.assertEqual(0x20, status_bytes[0], "Cooling status set")
        self.assertNotEqual(status_bytes[6:10], b"0028", "Temp != 4.0 C")

        # Hold for a while
        linkam.hold()
        linkam_device.process(30)
        status_bytes = linkam.get_status()
        self.assertEqual(0x50, status_bytes[0], "Manually holding")
        self.assertNotEqual(status_bytes[6:10], b"0028", "Temp != 4.0 C")

        # Cool some more
        linkam.cool()
        linkam_device.process(15)
        status_bytes = linkam.get_status()
        self.assertNotEqual(status_bytes[6:10], b"0028", "Temp != 4.0 C")

        # Hold again
        linkam.hold()
        linkam_device.process(30)
        status_bytes = linkam.get_status()
        self.assertEqual(0x50, status_bytes[0], "Manually holding")
        self.assertNotEqual(status_bytes[6:10], b"0028", "Temp != 4.0 C")

        # Finish cooling via heat command (should also work)
        linkam.heat()
        linkam_device.process(15)
        status_bytes = linkam.get_status()
        self.assertEqual(status_bytes[6:10], b"0028", "Temp == 4.0 C")

        # Make sure transitions to auto-holding
        linkam_device.process()
        status_bytes = linkam.get_status()
        self.assertEqual(0x30, status_bytes[0], "Auto-holding at limit")

    def test_pump_command(self):
        linkam = LinkamT95StreamInterface()
        linkam_device = SimulatedLinkamT95()
        linkam.device = linkam_device

        linkam_device.process()  # Initialize

        # Issue T command to get into stopped state
        linkam.get_status()
        linkam_device.process()

        # Set up to cool from 24.0 C to 4.0 C at 20.00 C/min
        linkam.set_rate("2000")
        linkam.set_limit("40")
        linkam.start()
        linkam_device.process()

        # Since the pump feature is not fully implemented,
        # we can only make sure all valid input is accepted
        assertRaisesNothing(self, linkam.pump_command, b"m0")  # Manual
        linkam_device.process()

        for int_value, char_value in enumerate(b"0123456789:;<=>?@ABCDEFGHIJKLMN"):
            assertRaisesNothing(
                self, linkam.pump_command, char_value
            )  # Characters mean speeds 0 - 30
            linkam_device.process()
            status_bytes = linkam.get_status()
            self.assertEqual(
                0x80 | int_value,
                status_bytes[2],
                "Verify Pump Status Byte reflects speed",
            )

        assertRaisesNothing(self, linkam.pump_command, b"a0")  # Auto
        linkam_device.process()



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
The statemachine module contains one of lewis' central parts, the cycle-based
:class:`StateMachine`. The module also contains classes that make it easier to define the
state machine (:class:`State`, :class:`Transition`). Despite its central nature, it's unlikely
to be used directly in client code for device simulations - these should be based on
:class:`StateMachineDevice`, which provides a more convenient interface for that purpose.
"""

from lewis.core.logging import has_log
from lewis.core.processor import CanProcess


class StateMachineException(Exception):
    """
    Classes in this module should only raise this type of Exception.
    """

    pass


class HasContext:
    """
    Mixin to provide a Context.

    Creates a `_context` member variable that can be assigned with :meth:`set_context`.

    Any state handler or transition callable that derives from this mixin will
    receive a context from its :class:`StateMachine` upon initialization (assuming the
    StateMachine was provided with a context itself).
    """

    def __init__(self) -> None:
        super(HasContext, self).__init__()
        self._context = None

    def set_context(self, new_context) -> None:
        """Assigns the new context to the member variable ``_context``."""
        self._context = new_context

        if hasattr(self, "_set_logging_context"):
            self._set_logging_context(self._context)


@has_log
class State(HasContext):
    """
    StateMachine state handler base class.

    Provides a way to implement StateMachine event handling behaviour using an
    object-oriented interface. Once the StateMachine is configured to do so, it
    will automatically invoke the events in this class when appropriate.

    To use this class, create a derived class and override any events that need
    custom behaviour. Device context is provided via :class:`HasContext` mixin.
    """

    def __init__(self) -> None:
        super(State, self).__init__()

    def on_entry(self, dt) -> None:
        """
        Handle entry event. Raised once, when this state is entered.

        :param dt: Delta T since last cycle.
        """
        pass

    def in_state(self, dt) -> None:
        """
        Handle in-state event.

        Raised repeatedly, once per cycle, while idling in this state. Exactly one
        in-state event occurs per cycle for every StateMachine. This is always the
        last event of the cycle.

        :param dt: Delta T since last cycle.
        """
        pass

    def on_exit(self, dt) -> None:
        """
        Handle exit event. Raised once, when this state is exited.

        :param dt: Delta T since last cycle.
        """
        pass


@has_log
class Transition(HasContext):
    """
    StateMachine transition condition base class.

    Provides a way to implement a transition that requires access to the device
    context. The device context is provided via :class:`HasContext` mixin, and can be
    accessed as `self._context`.

    To use this class, create a derived class and override the :meth:`__call__` attribute.
    """

    def __init__(self) -> None:
        super(Transition, self).__init__()

    def __call__(self):
        """
        This is invoked when the StateMachine wants to check whether this transition
        should occur. This happens on cycles when the StateMachine starts the cycle
        in the source state of this transition.

        If this call returns True, the StateMachine will transition to the destination
        state. Any remaining transition checks for the source state are not checked.

        :return: True or False / Should transition occur or not
        """
        return True


@has_log
class StateMachine(CanProcess):
    """
    Cycle based state machine.

    :param cfg: dict which contains state machine configuration.
    :param context: object which is assigned to State and Transition objects as their _context.

    The configuration dict may contain the following keys:

     - initial: Name of the initial state of this machine
     - states: [optional] Dict of custom state handlers
     - transitions: [optional] Dict of transitions in this state machine.

    State handlers may be given as a dict, list or State class:

     - dict: May contain keys 'on_entry', 'in_state' and 'on_exit'.
     - list: May contain up to 3 entries, above events in that order.
     - class: Should be an instance of a class that derives from State.

    In case of handlers being provided as a dict or a list, values should be callable
    and may take a single parameter: the Delta T since the last cycle.

    Transitions should be provided as a dict where:

     - Each key is a tuple of two values, the FROM and TO states respectively.
     - Each value is a callable transition condition that return True or False.

    Transition conditions are called once per cycle when in the FROM state. If one of
    the transition conditions returns True, the transition is executed that cycle. The
    remaining conditions aren't called.

    Consider using an OrderedDict if order matters.

    Only one transition may occur per cycle. Every cycle will, at the very least,
    trigger an in_state event against the current state.

    .. seealso:: See :meth:`~StateMachine.doProcess` for details.
    """

    def __init__(self, cfg, context=None) -> None:
        super(StateMachine, self).__init__()

        self._set_logging_context(context)

        self._state = None  # We start outside of any state, first cycle enters initial state
        self._handler = {}  # Nested dict mapping [state][event] = handler
        self._transition = {}  # Dict mapping [from_state] = [ (to_state, transition), ... ]
        self._prefix = {  # Default prefixes used when calling handler functions by name
            "on_entry": "_on_entry_",
            "in_state": "_in_state_",
            "on_exit": "_on_exit_",
        }

        # Specifying an initial state is not optional
        if "initial" not in cfg:
            raise StateMachineException(
                "StateMachine configuration must include " "'initial' to specify starting state."
            )
        self._initial = cfg["initial"]
        self._set_handlers(self._initial)

        self._setup_state_handlers(cfg.get("states", {}), context)
        self._setup_transition_handlers(cfg.get("transitions", {}), context)

    def _setup_state_handlers(self, state_handler_configuration, context) -> None:
        """
        This method constructs the state handlers from a user-provided dict.

        :param state_handler_configuration: Dictionary with state handler
        definitions.
        :param context: Context is provided to state handlers that inherit
        from HasContext.
        """
        for state_name, handlers in state_handler_configuration.items():
            if isinstance(handlers, HasContext):
                handlers.set_context(context)

            try:
                if isinstance(handlers, State):
                    self._set_handlers(
                        state_name,
                        handlers.on_entry,
                        handlers.in_state,
                        handlers.on_exit,
                    )
                elif isinstance(handlers, dict):
                    self._set_handlers(state_name, **handlers)
                elif hasattr(handlers, "__iter__"):
                    self._set_handlers(state_name, *handlers)
                else:
                    raise RuntimeError("Handler is not State, dict or __iter__.")
            except Exception:
                raise StateMachineException(
                    "Failed to parse state handlers for state '%s'. "
                    "Must be dict or iterable." % state_name
                )

    def _setup_transition_handlers(self, transition_handler_configuration, context) -> None:
        """
        This method constructs the transition handlers from a user-provided
        dict.

        :param transition_handler_configuration: Dictionary with transition
            handler definitions.
        :param context: Context is provided to transition handlers that inherit
        from HasContext.
        """
        for states, check_func in transition_handler_configuration.items():
            from_state, to_state = states

            # Set up default handlers if this state hasn't been mentioned before
            if from_state not in self._handler:
                self._set_handlers(from_state)
            if to_state not in self._handler:
                self._set_handlers(to_state)

            if isinstance(check_func, HasContext):
                check_func.set_context(context)

            # Set up the transition
            self._set_transition(from_state, to_state, check_func)

    @property
    def state(self):
        """Name of the current state."""
        return self._state

    def can(self, state):
        """
        Returns true if the transition to 'state' is allowed from the current state.

        :param state: State to check transition to
        :return: True if state is reachable from current
        """
        if self._state is None:
            return state == self._initial

        return state in (transition[0] for transition in self._transition[self._state])

    def bind_handlers_by_name(self, instance, override=False, prefix=None) -> None:
        """
        Auto-bind state handlers based on naming convention.

        :param instance: Target object instance to search for handlers and bind events to.
        :param override: If set to True, matching handlers will replace
                         previously registered handlers.
        :param prefix: Dict or list of prefixes to override defaults
                       (keys: on_entry, in_state, on_exit)

        This function enables automatically binding state handlers to events without having to
        specify them in the constructor. When called, this function searches `instance` for
        member functions that match the following patterns for all known states
        (states mentioned in 'states' or 'transitions' dicts of cfg):

            - ``instance._on_entry_[state]``
            - ``instance._in_state_[state]``
            - ``instance._on_exit_[state]``

        The default prefixes may be overridden using the prefix parameter. Supported keys are
        'on_entry', 'in_state', and 'on_exit'. Values should include any and
        all desired underscores.

        Matching functions are assigned as handlers to the corresponding state events,
        iff no handler was previously assigned to that event.

        If a state event already had a handler assigned (during construction or previous call
        to this function), no changes are made even if a matching function is found. To force
        previously assigned handlers to be overwritten, set the third parameter to True.
        This may be useful to implement inheritance-like specialization using multiple
        implementation classes but only one StateMachine instance.
        """
        if prefix is None:
            prefix = {}
        if not isinstance(prefix, dict) and hasattr(prefix, "__iter__"):
            prefix = dict(zip(["on_entry", "in_state", "on_exit"], prefix))

        # Merge prefix defaults with any provided prefixes
        prefix = dict(list(self._prefix.items()) + list(prefix.items()))

        # Bind handlers
        for state, handlers in self._handler.items():
            for event, handler in handlers.items():
                if handler is None or override:
                    named_handler = getattr(instance, prefix[event] + state, None)
                    if callable(named_handler):
                        self._handler[state][event] = named_handler

    def doProcess(self, dt) -> None:
        """
        Process a cycle of this state machine.

        :param dt: Delta T. "Time" passed since last cycle, passed on to event handlers.

        A cycle will perform at most one transition and exactly one in_state event.

        A transition will only occur if one of the transition condition functions leaving
        the current state returns True.

        When a transition occurs, the following events are raised:
         - on_exit_old_state()
         - on_entry_new_state()
         - in_state_new_state()

        The first cycle after init or reset will never call transition checks and, instead,
        always performs on_entry and in_state on the initial state.

        Whether a transition occurs or not, and regardless of any other circumstances, a
        cycle always ends by raising an in_state event on the current (potentially new)
        state.
        """
        # Initial transition on first cycle / after a reset()
        if self._state is None:
            self.log.debug('Entering initial state "%s"', self._initial)
            self._state = self._initial
            self._raise_event("on_entry", 0)
            self._raise_event("in_state", 0)
            return

        # General transition
        for target_state, check_func in self._transition.get(self._state, []):
            if check_func():
                self.log.debug("Transition triggered (%s -> %s)", self._state, target_state)
                self._raise_event("on_exit", dt)
                self._state = target_state
                self._raise_event("on_entry", dt)
                break

        # Always end with an in_state
        self._raise_event("in_state", dt)

    def reset(self) -> None:
        """
        Reset the state machine to before the first cycle. The next process() will
        enter the initial state.
        """
        self._state = None

    def _set_handlers(self, state, *args, **kwargs) -> None:
        """
        Add or update state handlers.

        :param state: Name of state to be added or updated
        :param on_entry: Handler for on_entry events. May be None, callable, or list of callables.
        :param in_state: Handler for in_state events. May be None, callable, or list of callables.
        :param on_exit: Handler for on_exit events. May be None, callable, or list of callables.

        Handlers may take up to one parameter (not counting self), delta T since last cycle,
        and should return nothing.

        When handlers are omitted or set to None, no event will be raised at all.
        """
        # Variable arguments for state handlers
        # Default to calling target.on_entry_state_name(), etc
        on_entry = args[0] if len(args) > 0 else kwargs.get("on_entry", None)
        in_state = args[1] if len(args) > 1 else kwargs.get("in_state", None)
        on_exit = args[2] if len(args) > 2 else kwargs.get("on_exit", None)

        self._handler[state] = {
            "on_entry": on_entry,
            "in_state": in_state,
            "on_exit": on_exit,
        }

    def _set_transition(self, from_state, to_state, transition_check) -> None:
        """
        Add or update a transition and its condition function.

        :param from_state: Name of state this transition leaves
        :param to_state: Name of state this transition enters
        :param transition_check: Callable condition under which this transition occurs.
                                 Should return True or False.

        The transition_check function should return True if the transition should occur.
        Otherwise, False.

        Transition condition functions should take no parameters (not counting self).
        """
        if not callable(transition_check):
            raise StateMachineException("Transition condition must be callable.")

        if from_state not in self._transition.keys():
            self._transition[from_state] = []

        # Remove previously added transition with same From -> To mapping
        try:
            del self._transition[from_state][
                [x[0] for x in self._transition[from_state]].index(to_state)
            ]
        except Exception:
            pass

        self._transition[from_state].append(
            (
                to_state,
                transition_check,
            )
        )

    def _raise_event(self, event, dt) -> None:
        """
        Invoke the given event name for the current state, passing dt as a parameter.

        :param event: Name of event to raise on current state.
        :param dt: Delta T since last cycle.
        """
        # May be None, function reference, or list of function refs
        self.log.debug("Processing state=%s, handler=%s", self._state, event)
        handlers = self._handler[self._state][event]

        if handlers is None:
            handlers = []

        if callable(handlers):
            handlers = [handlers]

        for handler in handlers:
            try:
                handler(dt)
            except TypeError:
                handler()



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest

from lewis.core.logging import has_log, root_logger_name


class TestHasLog(unittest.TestCase):
    def test_logger_name(self):
        @has_log
        class Foo:
            pass

        a = Foo()

        self.assertEqual(a.log.name, "{}.Foo".format(root_logger_name))

    def test_setting_context_changes_name(self):
        @has_log
        class Foo:
            pass

        a = Foo()
        self.assertEqual(a.log.name, "{}.Foo".format(root_logger_name))

        a._set_logging_context("some_context")
        self.assertEqual(a.log.name, "{}.some_context.Foo".format(root_logger_name))

        a._set_logging_context(None)
        self.assertEqual(a.log.name, "{}.Foo".format(root_logger_name))

    def test_decorate_function(self):
        @has_log
        def foo(bar):
            return bar

        self.assertTrue(hasattr(foo, "log"))
        self.assertEqual(foo.log.name, "{}.foo".format(root_logger_name))



# -*- coding: utf-8 -*-
#
# lewis documentation build configuration file, created by
# sphinx-quickstart on Wed Nov  9 16:42:53 2016.
import os
import sys
sys.path.insert(0, os.path.abspath("../lewis"))


# -- General configuration ------------------------------------------------
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    # and making summary tables at the top of API docs
    "sphinx.ext.autosummary",
    # This can parse google style docstrings
    "sphinx.ext.napoleon",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
]
templates_path = ["_templates"]
# General information about the project.
project = u"lewis"
language = 'en'
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
# -- Options for HTML output ---------------------------------------------
suppress_warnings =["docutils"]
html_theme = "sphinx_rtd_theme"
html_logo = "resources/logo/lewis-logo.png"
html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "ISISComputingGroup",  # Username
    "github_repo": "lewis",  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/doc/",  # Path in the checkout to the docs root
}

autoclass_content = "both"
myst_heading_anchors = 3

napoleon_google_docstring = False
napoleon_numpy_docstring = True



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import unittest
from unittest.mock import ANY, MagicMock, Mock, call, patch

from lewis.core.simulation import Simulation

from .utils import assertRaisesNothing


def set_simulation_running(environment):
    environment._running = True
    environment._started = True


class TestSimulation(unittest.TestCase):
    def setUp(self):
        # This makes sure that no actual sleeps happen during the tests.
        # The recipe is from the mock documentation.
        patcher = patch("lewis.core.simulation.sleep")
        self.addCleanup(patcher.stop)
        self.mock_sleep = patcher.start()

    @patch("lewis.core.simulation.seconds_since")
    def test_process_cycle_returns_elapsed_time(self, elapsed_seconds_mock):
        env = Simulation(device=Mock())

        # It doesn't matter what happens in the simulation cycle, here we
        # only care how long it took.
        with patch.object(env, "_process_simulation_cycle"):
            elapsed_seconds_mock.return_value = 0.5
            delta = env._process_cycle(0.0)

            elapsed_seconds_mock.assert_called_once_with(ANY)
            self.assertEqual(delta, 0.5)

    @patch("lewis.core.simulation.seconds_since")
    def test_process_cycle_changes_runtime_status(self, elapsed_seconds_mock):
        env = Simulation(device=Mock())

        with patch.object(env, "_process_simulation_cycle"):
            self.assertEqual(env.uptime, 0.0)

            set_simulation_running(env)

            elapsed_seconds_mock.return_value = 0.5
            env._process_cycle(0.0)

            self.assertEqual(env.uptime, 0.5)

    def test_pause_resume(self):
        env = Simulation(device=Mock())

        self.assertFalse(env.is_started)
        self.assertFalse(env.is_paused)

        # env is not running, so it can't be paused
        self.assertRaises(RuntimeError, env.pause)

        # Fake start of simulation, we don't need to care how this happened
        set_simulation_running(env)

        self.assertTrue(env.is_started)
        self.assertFalse(env.is_paused)

        assertRaisesNothing(self, env.pause)

        self.assertTrue(env.is_started)
        self.assertTrue(env.is_paused)

        assertRaisesNothing(self, env.resume)

        # now it's running, so it can't be resumed again
        self.assertRaises(RuntimeError, env.resume)

    def test_process_cycle_calls_sleep_if_paused(self):
        device_mock = Mock()
        env = Simulation(device=device_mock)
        set_simulation_running(env)
        env.pause()

        # simulation paused, device should not be called
        env._process_cycle(0.5)
        device_mock.assert_not_called()

        env.resume()
        # simulation is running now, device should be called
        env._process_cycle(0.5)

        device_mock.assert_has_calls([call.process(0.5)])

    def test_process_cycle_calls_process_simulation(self):
        device_mock = Mock()
        env = Simulation(device=device_mock)
        set_simulation_running(env)

        env._process_cycle(0.5)
        device_mock.assert_has_calls([call.process(0.5)])

        self.assertEqual(env.cycles, 1)
        self.assertEqual(env.runtime, 0.5)

    def test_process_simulation_cycle_applies_speed(self):
        device_mock = Mock()

        env = Simulation(device=device_mock)
        set_simulation_running(env)

        env.speed = 2.0
        env._process_cycle(0.5)

        device_mock.assert_has_calls([call.process(1.0)])

        self.assertEqual(env.cycles, 1)
        self.assertEqual(env.runtime, 1.0)

    def test_None_control_server_is_None(self):
        env = Simulation(device=Mock(), control_server=None)

        self.assertIsNone(env.control_server)

    def test_invalid_control_server_fails(self):
        self.assertRaises(Exception, Simulation, device=Mock(), control_server=5.0)
        self.assertRaises(Exception, Simulation, device=Mock(), control_server=3434)

        # With an arbitrary object it should also fail
        self.assertRaises(Exception, Simulation, device=Mock(), control_server=Mock())

        # The string must have two components separated by :
        self.assertRaises(Exception, Simulation, device=Mock(), control_server="a:b:c")

    @patch("lewis.core.simulation.ExposedObject")
    @patch("lewis.core.simulation.ControlServer")
    def test_construct_control_server(self, mock_control_server_type, exposed_object_mock):
        exposed_object_mock.return_value = "test"
        assertRaisesNothing(self, Simulation, device=Mock(), control_server="localhost:10000")

        mock_control_server_type.assert_called_once_with(
            {"device": "test", "simulation": "test", "interface": "test"},
            "localhost:10000",
        )

    def test_start_starts_control_server(self):
        env = Simulation(device=Mock())

        control_server_mock = Mock()
        env._control_server = control_server_mock

        def process_cycle_side_effect(delta):
            env.stop()

        env._process_cycle = Mock(side_effect=process_cycle_side_effect)
        env.start()

        control_server_mock.assert_has_calls([call.start_server()])

    def test_speed_range(self):
        env = Simulation(device=Mock())

        assertRaisesNothing(self, setattr, env, "speed", 3.0)
        self.assertEqual(env.speed, 3.0)

        assertRaisesNothing(self, setattr, env, "speed", 0.1)
        self.assertEqual(env.speed, 0.1)

        assertRaisesNothing(self, setattr, env, "speed", 0.0)
        self.assertEqual(env.speed, 0.0)

        self.assertRaises(ValueError, setattr, env, "speed", -0.5)

    def test_cycle_delay_range(self):
        env = Simulation(device=Mock())

        assertRaisesNothing(self, setattr, env, "cycle_delay", 0.2)
        self.assertEqual(env.cycle_delay, 0.2)

        assertRaisesNothing(self, setattr, env, "cycle_delay", 2.0)
        self.assertEqual(env.cycle_delay, 2.0)

        assertRaisesNothing(self, setattr, env, "cycle_delay", 0.0)
        self.assertEqual(env.cycle_delay, 0.0)

        self.assertRaises(ValueError, setattr, env, "cycle_delay", -4)

    def test_start_stop(self):
        env = Simulation(device=Mock())

        with patch.object(env, "_process_cycle", side_effect=lambda x: env.stop()) as mock_cycle:
            env.start()

            mock_cycle.assert_has_calls([call(0.0)])

    @patch("lewis.core.simulation.ExposedObject")
    @patch("lewis.core.simulation.ControlServer")
    def test_control_server_setter(self, control_server_mock, exposed_object_mock):
        # The return value (= instance of ControlServer) must be specified
        control_server_mock.return_value = Mock()
        exposed_object_mock.return_value = "test"

        env = Simulation(device=Mock())

        assertRaisesNothing(self, setattr, env, "control_server", "127.0.0.1:10001")
        control_server_mock.assert_called_once_with(
            {"device": "test", "simulation": "test", "interface": "test"},
            "127.0.0.1:10001",
        )

        control_server_mock.reset_mock()

        assertRaisesNothing(self, setattr, env, "control_server", None)
        self.assertIsNone(env.control_server)

        set_simulation_running(env)

        # Can set new control server even when simulation is running:
        assertRaisesNothing(self, setattr, env, "control_server", "127.0.0.1:10002")

        # The server is started automatically when the simulation is running
        control_server_mock.assert_called_once_with(
            {"device": "test", "simulation": "test", "interface": "test"},
            "127.0.0.1:10002",
        )

        # The instance must have one call to start_server
        control_server_mock.return_value.assert_has_calls([call.start_server()])

        # Can not replace control server when simulation is running
        self.assertRaises(RuntimeError, setattr, env, "control_server", "127.0.0.1:10003")

    def test_set_parameters(self):
        class TestDevice:
            foo = 10
            bar = "str"

            def baz(self):
                pass

        dev = TestDevice()

        sim = Simulation(device=dev)

        assertRaisesNothing(self, sim.set_device_parameters, {"foo": 5, "bar": "test"})

        self.assertEqual(dev.foo, 5)
        self.assertEqual(dev.bar, "test")

        self.assertRaises(RuntimeError, sim.set_device_parameters, {"not_existing": 45})
        self.assertRaises(RuntimeError, sim.set_device_parameters, {"baz": 4})

    def test_setups_empty(self):
        sim = Simulation(device=Mock(), device_builder=None)

        self.assertEqual(sim.setups, [])

    def test_setups(self):
        class MockBuilder:
            setups = {"foo": 1, "bar": 2}

        sim = Simulation(device=Mock(), device_builder=MockBuilder())

        self.assertEqual(set(sim.setups), {"foo", "bar"})

    def test_switch_setup(self):
        class MockBuilder:
            setups = {"foo": None}

            def create_device(self, setup):
                if setup == "foo":
                    return setup

                raise RuntimeError("Error")

        adapter_mock = MagicMock()
        sim = Simulation(device=Mock(), adapters=adapter_mock, device_builder=MockBuilder())

        sim.switch_setup("foo")

        self.assertEqual(sim._device, "foo")
        self.assertRaises(RuntimeError, sim.switch_setup, "bar")



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import socket
import unittest
from unittest.mock import Mock, call, patch

import zmq

from lewis.core.control_server import (
    ControlServer,
    ExposedObject,
    ExposedObjectCollection,
)
from lewis.core.exceptions import LewisException

from .utils import assertRaisesNothing


class DummyObject:
    a = 10
    b = 20

    def __init__(self):
        self.getTest = Mock()
        self.setTest = Mock()


class DummyObjectChild(DummyObject):
    c = 30


class TestRPCObject(unittest.TestCase):
    def test_all_methods_exposed(self):
        rpc_object = ExposedObject(DummyObject())

        expected_methods = [
            ":api",
            "a:get",
            "a:set",
            "b:get",
            "b:set",
            "getTest",
            "setTest",
        ]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_select_methods_exposed(self):
        rpc_object = ExposedObject(DummyObject(), ("a", "getTest"))

        expected_methods = [":api", "a:get", "a:set", "getTest"]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_excluded_methods_not_exposed(self):
        rpc_object = ExposedObject(DummyObject(), exclude=("a", "setTest"))

        expected_methods = [":api", "b:get", "b:set", "getTest"]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_selected_and_excluded_methods(self):
        rpc_object = ExposedObject(DummyObject(), members=("a", "getTest"), exclude=("a"))

        expected_methods = [":api", "getTest"]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_inherited_not_exposed(self):
        rpc_object = ExposedObject(DummyObjectChild(), members=("a", "c"), exclude_inherited=True)

        expected_methods = [":api", "c:get", "c:set"]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_inherited_exposed(self):
        rpc_object = ExposedObject(DummyObjectChild(), members=("a", "c"))

        expected_methods = [":api", "a:get", "a:set", "c:get", "c:set"]
        self.assertEqual(len(rpc_object), len(expected_methods))

        for method in expected_methods:
            self.assertTrue(method in rpc_object)

    def test_invalid_method_raises(self):
        self.assertRaises(AttributeError, ExposedObject, DummyObject(), ("nonExisting",))

    def test_attribute_wrapper_gets_value(self):
        obj = DummyObject()
        obj.a = 233

        rpc_object = ExposedObject(obj)
        self.assertEqual(rpc_object["a:get"](), obj.a)

    def test_attribute_wrapper_sets_value(self):
        obj = DummyObject()
        obj.a = 233

        rpc_object = ExposedObject(obj)

        self.assertEqual(obj.a, 233)
        rpc_object["a:set"](20)
        self.assertEqual(obj.a, 20)

    def test_attribute_wrapper_argument_number(self):
        rpc_object = ExposedObject(DummyObject())

        self.assertRaises(TypeError, rpc_object["a:get"], 20)
        self.assertRaises(TypeError, rpc_object["a:set"])
        self.assertRaises(TypeError, rpc_object["a:set"], 40, 30)

    def test_method_wrapper_calls(self):
        obj = DummyObject()
        rpc_object = ExposedObject(obj)

        rpc_object["getTest"](45, 56)

        obj.getTest.assert_called_with(45, 56)

    def test_get_api(self):
        obj = DummyObject()
        rpc_object = ExposedObject(obj, ["a"])
        api = rpc_object.get_api()

        self.assertTrue("class" in api)
        self.assertEqual(api["class"], type(obj).__name__)

        self.assertTrue("methods" in api)
        self.assertEqual(set(api["methods"]), {":api", "a:set", "a:get"})

    def test_lock_is_used_if_supplied(self):
        mock_lock = Mock()
        mock_lock.__enter__ = Mock()
        mock_lock.__exit__ = Mock()

        obj = DummyObject()
        exposed_object = ExposedObject(obj, ["a"], lock=mock_lock)

        self.assertEqual(exposed_object["a:get"](), obj.a)

        mock_lock.__enter__.assert_called_once()
        mock_lock.__exit__.assert_called_once()


class TestExposedObjectCollection(unittest.TestCase):
    def test_empty_initialization(self):
        exposed_objects = ExposedObjectCollection(named_objects={})
        self.assertEqual(set(exposed_objects), {":api", "get_objects"})

        self.assertEqual(len(exposed_objects.get_objects()), 0)
        self.assertEqual(exposed_objects["get_objects"](), exposed_objects.get_objects())

    def test_api(self):
        exposed_objects = ExposedObjectCollection(named_objects={})

        api = exposed_objects[":api"]()
        self.assertTrue("class" in api)
        self.assertEqual(api["class"], "ExposedObjectCollection")
        self.assertTrue("methods" in api)
        self.assertEqual(set(api["methods"]), {"get_objects", ":api"})

    def test_add_plain_object(self):
        exposed_objects = ExposedObjectCollection({})
        obj = DummyObject()

        assertRaisesNothing(self, exposed_objects.add_object, obj, "testObject")

        # There should be :api, get_objects, testObject:api, testObject.a:get, testObject.a:set,
        # testObject.b:get, testObject.b:set, testObject.getTest, testObject.setTest
        self.assertEqual(len(exposed_objects), 9)

        exposed_objects["testObject.getTest"](34, 55)
        obj.getTest.assert_called_once_with(34, 55)

    def test_add_exposed_object(self):
        exposed_objects = ExposedObjectCollection({})
        obj = DummyObject()

        assertRaisesNothing(
            self,
            exposed_objects.add_object,
            ExposedObject(obj, ("setTest", "getTest")),
            "testObject",
        )
        exposed_objects["testObject.getTest"](41, 11)
        obj.getTest.assert_called_once_with(41, 11)

    def test_nested_collections(self):
        obj = DummyObject()
        exposed_objects = ExposedObjectCollection(
            {"container": ExposedObjectCollection({"test": obj})}
        )

        exposed_objects["container.test.getTest"](454, 43)
        obj.getTest.assert_called_once_with(454, 43)

    def test_duplicate_name_raises(self):
        exposed_objects = ExposedObjectCollection({})
        exposed_objects.add_object(DummyObject(), "testObject")

        self.assertRaises(RuntimeError, exposed_objects.add_object, DummyObject(), "testObject")

    def test_remove_object(self):
        exposed_objects = ExposedObjectCollection({})
        own_functions_count = len(exposed_objects)

        obj = DummyObject()
        exposed_objects.add_object(obj, "testObject")

        self.assertListEqual(exposed_objects.get_objects(), ["testObject"])
        assertRaisesNothing(self, exposed_objects.remove_object, "testObject")
        self.assertEqual(len(exposed_objects), own_functions_count)

        self.assertRaises(RuntimeError, exposed_objects.remove_object, "does_not_exist")


class TestControlServer(unittest.TestCase):
    @patch("zmq.Context")
    def test_connection(self, mock_context):
        cs = ControlServer(None, connection_string="127.0.0.1:10001")
        cs.start_server()

        mock_context.assert_has_calls(
            [
                call(),
                call().socket(zmq.REP),
                call().socket().setsockopt(zmq.RCVTIMEO, 100),
                call().socket().bind("tcp://127.0.0.1:10001"),
            ]
        )

    @patch("zmq.Context")
    def test_server_can_only_be_started_once(self, mock_context):
        server = ControlServer(None, connection_string="127.0.0.1:10000")
        server.start_server()
        server.start_server()

        mock_context.assert_has_calls(
            [
                call(),
                call().socket(zmq.REP),
                call().socket().setsockopt(zmq.RCVTIMEO, 100),
                call().socket().bind("tcp://127.0.0.1:10000"),
            ]
        )

    def test_process_raises_if_not_started(self):
        server = ControlServer(None, connection_string="127.0.0.1:10000")

        self.assertRaises(Exception, server.process)

    def test_process_does_not_block(self):
        mock_socket = Mock()
        mock_socket.recv_unicode.side_effect = zmq.Again()

        server = ControlServer(None, connection_string="127.0.0.1:10000")
        server._socket = mock_socket
        assertRaisesNothing(self, server.process)

        mock_socket.recv_unicode.assert_has_calls([call(flags=zmq.NOBLOCK)])

    def test_exposed_object_is_exposed_directly(self):
        mock_collection = Mock(spec=ExposedObject)

        server = ControlServer(object_map=mock_collection, connection_string="127.0.0.1:10000")
        self.assertEqual(server.exposed_object, mock_collection)

    @patch("lewis.core.control_server.ExposedObjectCollection")
    def test_exposed_object_collection_is_constructed(self, exposed_object_mock):
        ControlServer(object_map="test", connection_string="127.0.0.1:10000")

        exposed_object_mock.assert_called_once_with("test")

    @patch("zmq.Context")
    def test_is_running(self, mock_context):
        server = ControlServer(None, connection_string="127.0.0.1:10000")
        self.assertFalse(server.is_running)
        server.start_server()
        self.assertTrue(server.is_running)

    @patch("lewis.core.control_server.socket.gethostbyname")
    def test_invalid_hostname_raises_LewisException(self, gethostbyname_mock):
        def raise_exception(self):
            raise socket.gaierror

        gethostbyname_mock.side_effect = raise_exception

        self.assertRaises(
            LewisException,
            ControlServer,
            object_map=None,
            connection_string="some_invalid_host.local:10000",
        )

        gethostbyname_mock.assert_called_once_with("some_invalid_host.local")

    def test_localhost_does_not_raise_socket_error(self):
        assertRaisesNothing(
            self, ControlServer, object_map=None, connection_string="localhost:10000"
        )
